from unittest import TestCase

import orjson as json

from panther import Panther
from panther.app import API, GenericAPI
from panther.request import Request
from panther.response import Response
from panther.test import APIClient


@API()
async def request_path(request: Request):
    return request.path


@API()
async def request_client(request: Request):
    return request.client


@API()
async def request_query_params(request: Request):
    return request.query_params


@API()
async def request_data(request: Request):
    return request.data


@API()
async def request_path_variables(name: str, age: int, is_alive: bool):
    return {'name': name, 'age': age, 'is_alive': is_alive}


@API()
async def request_header(request: Request):
    return request.headers.__dict__


@API()
async def request_header_by_attr(request: Request):
    return request.headers.authorization


@API()
async def request_header_by_item(request: Request):
    return request.headers['Authorization']


# # # Methods
class AllMethods(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()

    def post(self, *args, **kwargs):
        return Response()

    def put(self, *args, **kwargs):
        return Response()

    def patch(self, *args, **kwargs):
        return Response()

    def delete(self, *args, **kwargs):
        return Response()


@API()
async def all_methods():
    return Response()


class GetMethod(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()


@API(methods=['GET'])
async def get_method():
    return Response()


class PostMethod(GenericAPI):
    def post(self, *args, **kwargs):
        return Response()


@API(methods=['POST'])
async def post_method():
    return Response()


class PutMethod(GenericAPI):
    def put(self, *args, **kwargs):
        return Response()


@API(methods=['PUT'])
async def put_method():
    return Response()


class PatchMethod(GenericAPI):
    def patch(self, *args, **kwargs):
        return Response()


@API(methods=['PATCH'])
async def patch_method():
    return Response()


class DeleteMethod(GenericAPI):
    def delete(self, *args, **kwargs):
        return Response()


@API(methods=['DELETE'])
async def delete_method():
    return Response()


class GetPostPatchMethods(GenericAPI):
    def get(self, *args, **kwargs):
        return Response()

    def post(self, *args, **kwargs):
        return Response()

    def patch(self, *args, **kwargs):
        return Response()


@API(methods=['GET', 'POST', 'PATCH'])
async def get_post_patch_methods():
    return Response()


urls = {
    'path': request_path,
    'client': request_client,
    'query-params': request_query_params,
    'data': request_data,
    'path/<name>/variable/<age>/<is_alive>/': request_path_variables,

    'header': request_header,
    'header-attr': request_header_by_attr,
    'header-item': request_header_by_item,

    'all-func': all_methods,
    'all-class': AllMethods,
    'get-func': get_method,
    'get-class': GetMethod,
    'post-func': post_method,
    'post-class': PostMethod,
    'put-func': put_method,
    'put-class': PutMethod,
    'patch-func': patch_method,
    'patch-class': PatchMethod,
    'delete-func': delete_method,
    'delete-class': DeleteMethod,
    'get-post-patch-func': get_post_patch_methods,
    'get-post-patch-class': GetPostPatchMethods,
}


class TestRequest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def test_path(self):
        res = self.client.get('path/')
        assert res.status_code == 200
        assert res.data == '/path/'

    def test_client(self):
        res = self.client.get('client/')
        assert res.status_code == 200
        assert res.data == ['127.0.0.1', 8585]

    def test_query_params(self):
        res = self.client.get(
            'query-params/',
            query_params={'my': 'name', 'is': 'ali', 'how': 'are'},
        )
        assert res.status_code == 200
        assert res.data == {'my': 'name', 'is': 'ali', 'how': 'are'}

    def test_data(self):
        payload = {'detail': 'ok'}
        res = self.client.post('data/', payload=json.dumps(payload))
        assert res.status_code == 200
        assert res.data == payload

    def test_path_variables(self):
        res = self.client.post('path/Ali/variable/27/true/')
        expected_response = {'name': 'Ali', 'age': 27, 'is_alive': True}
        assert res.status_code == 200
        assert res.data == expected_response

    # # # Headers
    def test_headers_none(self):
        res = self.client.get('header')
        expected_headers = {}
        assert res.data == expected_headers

    def test_headers_content_type(self):
        res = self.client.post('header')
        expected_headers = {'content-type': 'application/json'}
        assert res.data == expected_headers

    def test_headers_full_items(self):
        headers = {
            'User-Agent': 'PostmanRuntime/7.36.0',
            'Accept': '*/*',
            'Cache-Control': 'no-cache',
            'Host': '127.0.0.1:8000',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Content-Length': 546,
        }
        res = self.client.post('header', headers=headers)
        expected_headers = {
            'content-type': 'application/json',
            'User-Agent': 'PostmanRuntime/7.36.0',
            'Accept': '*/*',
            'Cache-Control': 'no-cache',
            'Host': '127.0.0.1:8000',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Content-Length': '546',
        }
        assert res.data == expected_headers

    def test_headers_unknown_items(self):
        headers = {
            'Header1': 'PostmanRuntime/7.36.0',
            'Header2': '*/*',
        }
        res = self.client.post('header', headers=headers)
        expected_headers = {
            'content-type': 'application/json',
            'Header1': 'PostmanRuntime/7.36.0',
            'Header2': '*/*',
        }
        assert res.data == expected_headers

    def test_headers_authorization_by_getattr(self):
        headers = {
            'Authorization': 'Token xxx',
        }
        res = self.client.post('header-attr', headers=headers)
        assert res.data == 'Token xxx'

    def test_headers_authorization_by_getitem(self):
        headers = {
            'Authorization': 'Token xxx',
        }
        res = self.client.post('header-item', headers=headers)
        assert res.data == 'Token xxx'

    # # # Methods
    def test_method_all(self):
        res_func = self.client.get('all-func/')
        res_class = self.client.get('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.post('all-func/')
        res_class = self.client.post('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.put('all-func/')
        res_class = self.client.put('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.patch('all-func/')
        res_class = self.client.patch('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.delete('all-func/')
        res_class = self.client.delete('all-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

    def test_method_get(self):
        res_func = self.client.get('get-func/')
        res_class = self.client.get('get-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.post('get-func/')
        res_class = self.client.post('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.put('get-func/')
        res_class = self.client.put('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.patch('get-func/')
        res_class = self.client.patch('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.delete('get-func/')
        res_class = self.client.delete('get-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    def test_method_post(self):
        res_func = self.client.get('post-func/')
        res_class = self.client.get('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.post('post-func/')
        res_class = self.client.post('post-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.put('post-func/')
        res_class = self.client.put('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.patch('post-func/')
        res_class = self.client.patch('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.delete('post-func/')
        res_class = self.client.delete('post-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    def test_method_put(self):
        res_func = self.client.get('put-func/')
        res_class = self.client.get('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.post('put-func/')
        res_class = self.client.post('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.put('put-func/')
        res_class = self.client.put('put-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.patch('put-func/')
        res_class = self.client.patch('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.delete('put-func/')
        res_class = self.client.delete('put-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    def test_method_patch(self):
        res_func = self.client.get('patch-func/')
        res_class = self.client.get('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.post('patch-func/')
        res_class = self.client.post('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.put('patch-func/')
        res_class = self.client.put('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.patch('patch-func/')
        res_class = self.client.patch('patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.delete('patch-func/')
        res_class = self.client.delete('patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

    def test_method_delete(self):
        res_func = self.client.get('delete-func/')
        res_class = self.client.get('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.post('delete-func/')
        res_class = self.client.post('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.put('delete-func/')
        res_class = self.client.put('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.patch('delete-func/')
        res_class = self.client.patch('delete-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.delete('delete-func/')
        res_class = self.client.delete('delete-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

    def test_method_get_post_patch(self):
        res_func = self.client.get('get-post-patch-func/')
        res_class = self.client.get('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.post('get-post-patch-func/')
        res_class = self.client.post('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.put('get-post-patch-func/')
        res_class = self.client.put('get-post-patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405

        res_func = self.client.patch('get-post-patch-func/')
        res_class = self.client.patch('get-post-patch-class/')
        assert res_func.status_code == 200
        assert res_class.status_code == 200

        res_func = self.client.delete('get-post-patch-func/')
        res_class = self.client.delete('get-post-patch-class/')
        assert res_func.status_code == 405
        assert res_class.status_code == 405



import typing

from pydantic import create_model, BaseModel
from pydantic.fields import FieldInfo
from pydantic_core._pydantic_core import PydanticUndefined

from panther.db import Model


class MetaModelSerializer:
    KNOWN_CONFIGS = ['model', 'fields', 'required_fields']

    def __new__(
            cls,
            cls_name: str,
            bases: tuple[type[typing.Any], ...],
            namespace: dict[str, typing.Any],
            **kwargs
    ):
        if cls_name == 'ModelSerializer':
            cls.model_serializer = type(cls_name, (), namespace)
            return super().__new__(cls)

        # 1. Initial Check
        cls.check_config(cls_name=cls_name, namespace=namespace)
        config = namespace.pop('Config')

        # 2. Collect `Fields`
        field_definitions = cls.collect_fields(config=config, namespace=namespace)

        # 3. Collect `pydantic.model_config`
        model_config = cls.collect_model_config(config=config, namespace=namespace)
        namespace |= {'model_config': model_config}

        # 4. Create a serializer
        return create_model(
            __model_name=cls_name,
            __module__=namespace['__module__'],
            __validators__=namespace,
            __base__=(cls.model_serializer, BaseModel),
            __doc__=namespace.get('__doc__'),
            model=(typing.ClassVar, config.model),
            **field_definitions
        )

    @classmethod
    def check_config(cls, cls_name: str, namespace: dict) -> None:
        module = namespace['__module__']
        address = f'{module}.{cls_name}'

        # Check `Config`
        if (config := namespace.get('Config')) is None:
            msg = f'`class Config` is required in {address}.'
            raise AttributeError(msg) from None

        # Check `model`
        if (model := getattr(config, 'model', None)) is None:
            msg = f'`{cls_name}.Config.model` is required.'
            raise AttributeError(msg) from None

        # Check `model` type
        try:
            if not issubclass(model, Model):
                msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model`.'
                raise AttributeError(msg) from None
        except TypeError:
            msg = f'`{cls_name}.Config.model` is not subclass of `panther.db.Model`.'
            raise AttributeError(msg) from None

        # Check `fields`
        if (fields := getattr(config, 'fields', None)) is None:
            msg = f'`{cls_name}.Config.fields` is required.'
            raise AttributeError(msg) from None

        for field_name in fields:
            if field_name not in model.model_fields:
                msg = f'`{cls_name}.Config.fields.{field_name}` is not valid.'
                raise AttributeError(msg) from None

        # Check `required_fields`
        if not hasattr(config, 'required_fields'):
            config.required_fields = []

        for required in config.required_fields:
            if required not in config.fields:
                msg = f'`{cls_name}.Config.required_fields.{required}` should be in `Config.fields` too.'
                raise AttributeError(msg) from None

    @classmethod
    def collect_fields(cls, config: typing.Callable, namespace: dict) -> dict:
        field_definitions = {}

        # Define `fields`
        for field_name in config.fields:
            field_definitions[field_name] = (
                config.model.model_fields[field_name].annotation,
                config.model.model_fields[field_name]
            )

        # Apply `required_fields`
        for required in config.required_fields:
            field_definitions[required][1].default = PydanticUndefined

        # Collect and Override `Class Fields`
        for key, value in namespace.pop('__annotations__', {}).items():
            field_info = namespace.pop(key, FieldInfo(required=True))
            field_info.annotation = value
            field_definitions[key] = (value, field_info)

        return field_definitions

    @classmethod
    def collect_model_config(cls, config: typing.Callable, namespace: dict) -> dict:
        return {
            attr: getattr(config, attr) for attr in dir(config)
            if not attr.startswith('__') and attr not in cls.KNOWN_CONFIGS
        } | namespace.pop('model_config', {})


class ModelSerializer(metaclass=MetaModelSerializer):
    def create(self) -> type[Model]:
        return self.model.insert_one(self.model_dump())



import typing
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Callable

from pydantic._internal._model_construction import ModelMetaclass

from panther.throttling import Throttling
from panther.utils import Singleton


class JWTConfig:
    def __init__(
            self,
            key: str,
            algorithm: str = 'HS256',
            life_time: timedelta | int = timedelta(days=1),
            refresh_life_time: timedelta | int | None = None,
    ):
        self.key = key
        self.algorithm = algorithm
        self.life_time = life_time.total_seconds() if isinstance(life_time, timedelta) else life_time

        if refresh_life_time:
            if isinstance(refresh_life_time, timedelta):
                self.refresh_life_time = refresh_life_time.total_seconds()
            else:
                self.refresh_life_time = refresh_life_time
        else:
            self.refresh_life_time = self.life_time * 2


class QueryObservable:
    observers = []

    @classmethod
    def observe(cls, observer):
        cls.observers.append(observer)

    @classmethod
    def update(cls):
        for observer in cls.observers:
            observer._reload_bases(parent=config.query_engine)


@dataclass
class Config(Singleton):
    base_dir: Path
    monitoring: bool
    log_queries: bool
    default_cache_exp: timedelta | None
    throttling: Throttling | None
    secret_key: bytes | None
    http_middlewares: list
    ws_middlewares: list
    reversed_http_middlewares: list
    reversed_ws_middlewares: list
    user_model: ModelMetaclass | None
    authentication: ModelMetaclass | None
    jwt_config: JWTConfig | None
    models: list[dict]
    flat_urls: dict
    urls: dict
    query_engine: typing.Callable | None
    websocket_connections: typing.Callable | None
    background_tasks: bool
    has_ws: bool
    startup: Callable | None
    shutdown: Callable | None
    auto_reformat: bool
    pantherdb_encryption: bool

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'query_engine':
            QueryObservable.update()

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, item):
        return getattr(self, item)


config = Config(
    base_dir=Path(),
    monitoring=False,
    log_queries=False,
    default_cache_exp=None,
    throttling=None,
    secret_key=None,
    http_middlewares=[],
    ws_middlewares=[],
    reversed_http_middlewares=[],
    reversed_ws_middlewares=[],
    user_model=None,
    authentication=None,
    jwt_config=None,
    models=[],
    flat_urls={},
    urls={},
    query_engine=None,
    websocket_connections=None,
    background_tasks=False,
    has_ws=False,
    startup=None,
    shutdown=None,
    auto_reformat=False,
    pantherdb_encryption=False,
)



import os
from datetime import datetime, timedelta
from pathlib import Path
from unittest import TestCase

import panther.utils
from panther import Panther
from panther.middlewares import BaseMiddleware
from panther.utils import generate_hash_value_from_string, load_env, round_datetime, encrypt_password


class TestLoadEnvFile(TestCase):
    file_path = 'tests/env-test'

    is_active = True
    db_host = '127.0.0.1'
    db_port = 27017

    def tearDown(self) -> None:
        Path(self.file_path).unlink(missing_ok=True)

    def _create_env_file(self, file_data):
        with open(self.file_path, 'w') as file:
            file.write(file_data)

    def test_load_env_invalid_file(self):
        with self.assertLogs(level='ERROR') as captured:
            load_env('fake.file')

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '"fake.file" is not valid file for load_env()'

    def test_load_env_double_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = "{self.is_active}"
DB_HOST = "{self.db_host}"
DB_PORT = "{self.db_port}"
        """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_single_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_no_quote(self):
        self._create_env_file(f"""
IS_ACTIVE = {self.is_active}
DB_HOST = {self.db_host}
DB_PORT = {self.db_port}

                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_no_space(self):
        self._create_env_file(f"""
IS_ACTIVE={self.is_active}
DB_HOST={self.db_host}
DB_PORT={self.db_port}
                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_not_striped(self):
        self._create_env_file(f"""
        IS_ACTIVE = {self.is_active}
        DB_HOST = {self.db_host}
        DB_PORT = {self.db_port}
                    """)

        variables = load_env(self.file_path)
        self.assertEqual(variables['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(variables['DB_HOST'], self.db_host)
        self.assertEqual(variables['DB_PORT'], str(self.db_port))

    def test_load_env_and_read_from_system_env(self):
        self._create_env_file(f"""
IS_ACTIVE = '{self.is_active}'
DB_HOST = '{self.db_host}'
DB_PORT = '{self.db_port}'
                """)

        load_env(self.file_path)
        self.assertEqual(os.environ['IS_ACTIVE'] == 'True', self.is_active)
        self.assertEqual(os.environ['DB_HOST'], self.db_host)
        self.assertEqual(os.environ['DB_PORT'], str(self.db_port))


class TestUtilFunctions(TestCase):

    def test_round_datetime_second_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_second_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=35)
        _delta = timedelta(seconds=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=40)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_minute_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=15, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_minute_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(minutes=20)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12, minute=20)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_hour_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=12, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_hour_2(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(hours=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=25, hour=12)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_day_1(self):
        _datetime = datetime(year=1997, month=12, day=25, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_round_datetime_day_2(self):
        _datetime = datetime(year=1997, month=12, day=22, hour=10, minute=10, second=30)
        _delta = timedelta(days=5)
        rounded_datetime = round_datetime(_datetime, _delta)

        expected_datetime = datetime(year=1997, month=12, day=23)
        self.assertEqual(rounded_datetime, expected_datetime)

    def test_generate_hash_value_from_string(self):
        text = 'Hello World'
        hashed_1 = generate_hash_value_from_string(text)
        hashed_2 = generate_hash_value_from_string(text)

        self.assertEqual(hashed_1, hashed_2)
        self.assertNotEqual(text, hashed_1)

    def test_encrypt_password(self):
        password = 'Password'
        encrypted = encrypt_password(password=password)
        assert password != encrypted


class TestLoadConfigs(TestCase):
    def test_urls_not_found(self):
        global URLs
        URLs = None

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': is required."

    def test_urls_cant_be_dict(self):
        global URLs
        URLs = {}

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        msg = (
            "Invalid 'URLs': can't be 'dict', you may want to pass it's value directly to Panther(). "
            "Example: Panther(..., urls=...)"
        )
        assert captured.records[0].getMessage() == msg

    def test_urls_not_string(self):
        global URLs
        URLs = True

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': should be dotted string."

    def test_urls_invalid_target(self):
        global URLs
        URLs = 'tests.test_utils.TestLoadConfigs'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': should point to a dict."

    def test_urls_invalid_module_path(self):
        global URLs
        URLs = 'fake.module'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__)
            except SystemExit:
                assert True
            else:
                assert False

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'URLs': No module named 'fake'"

    def test_middlewares_invalid_path(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('fake.module', {})
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': fake.module is not a valid middleware path"

    def test_middlewares_invalid_structure(self):
        global MIDDLEWARES
        MIDDLEWARES = ['fake.module']

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': fake.module should have 2 part: (path, kwargs)"

    def test_middlewares_too_many_args(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('fake.module', 1, 2)
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': ('fake.module', 1, 2) too many arguments"

    def test_middlewares_without_args(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('tests.test_utils.CorrectTestMiddleware', )
        ]

        with self.assertNoLogs(level='ERROR'):
            Panther(name=__name__, configs=__name__, urls={})

        MIDDLEWARES = []

    def test_middlewares_invalid_middleware_parent(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            ('tests.test_utils.TestMiddleware', )
        ]

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == "Invalid 'MIDDLEWARES': is not a sub class of BaseMiddleware"

    def test_jwt_auth_without_secret_key(self):
        global AUTHENTICATION
        AUTHENTICATION = 'panther.authentications.JWTAuthentication'

        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                AUTHENTICATION = None

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '"SECRET_KEY" is required when using "JWTAuthentication"'

    def test_jwt_auth_with_secret_key(self):
        global AUTHENTICATION, SECRET_KEY
        AUTHENTICATION = 'panther.authentications.JWTAuthentication'
        SECRET_KEY = panther.utils.generate_secret_key()

        with self.assertNoLogs(level='ERROR'):
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert False
            else:
                assert True
            finally:
                AUTHENTICATION = None
                SECRET_KEY = None


class CorrectTestMiddleware(BaseMiddleware):
    pass


class TestMiddleware:
    pass



from collections import namedtuple
from collections.abc import Callable


class Headers:
    accept: str
    accept_encoding: str
    accept_language: str
    authorization: str
    cache_control: str
    connection: str
    content_length: str
    content_type: str
    host: str
    origin: str
    pragma: str
    referer: str
    sec_fetch_dest: str
    sec_fetch_mode: str
    sec_fetch_site: str
    user_agent: str

    upgrade: str
    sec_websocket_version: str
    sec_websocket_key: str

    def __init__(self, headers):
        self.__headers = headers
        self.__pythonic_headers = {k.lower().replace('-', '_'): v for k, v in headers.items()}

    def __getattr__(self, item: str):
        if result := self.__pythonic_headers.get(item):
            return result
        else:
            return self.__headers.get(item)

    def __getitem__(self, item: str):
        if result := self.__headers.get(item):
            return result
        else:
            return self.__pythonic_headers.get(item)

    def __str__(self):
        items = ', '.join(f'{k}={v}' for k, v in self.__headers.items())
        return f'Headers({items})'

    __repr__ = __str__

    @property
    def __dict__(self):
        return self.__headers


Address = namedtuple('Client', ['ip', 'port'])


class BaseRequest:
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        self.scope = scope
        self.asgi_send = send
        self.asgi_receive = receive
        self._data = ...
        self._validated_data = None
        self._user = None
        self._headers: Headers | None = None
        self._params: dict | None = None

    @property
    def headers(self) -> Headers:
        if self._headers is None:
            _headers = {header[0].decode('utf-8'): header[1].decode('utf-8') for header in self.scope['headers']}
            self._headers = Headers(_headers)
        return self._headers

    @property
    def query_params(self) -> dict:
        if self._params is None:
            self._params = {}
            if (query_string := self.scope['query_string']) != b'':
                query_string = query_string.decode('utf-8').split('&')
                for param in query_string:
                    k, *_, v = param.split('=')
                    self._params[k] = v
        return self._params

    @property
    def path(self) -> str:
        return self.scope['path']

    @property
    def server(self) -> Address:
        return Address(*self.scope['server'])

    @property
    def client(self) -> Address:
        return Address(*self.scope['client'])

    @property
    def http_version(self) -> str:
        return self.scope['http_version']

    @property
    def scheme(self) -> str:
        return self.scope['scheme']

    @property
    def user(self):
        return self._user

    def set_user(self, user) -> None:
        self._user = user



import asyncio
import importlib
import logging
import re
import subprocess
import types
from collections.abc import Callable
from traceback import TracebackException
from uuid import uuid4

import orjson as json

from panther import status
from panther.exceptions import PantherException
from panther.file_handler import File

logger = logging.getLogger('panther')


async def _http_response_start(send: Callable, /, headers: dict, status_code: int) -> None:
    bytes_headers = [[k.encode(), str(v).encode()] for k, v in (headers or {}).items()]
    await send({
        'type': 'http.response.start',
        'status': status_code,
        'headers': bytes_headers,
    })


async def _http_response_body(send: Callable, /, body: bytes | None = None) -> None:
    if body is None:
        await send({'type': 'http.response.body'})
    else:
        await send({'type': 'http.response.body', 'body': body})


async def http_response(
        send: Callable,
        /,
        *,
        status_code: int,
        monitoring=None,  # type: MonitoringMiddleware | None
        headers: dict | None = None,
        body: bytes | None = None,
        exception: bool = False,
) -> None:
    if exception:
        body = json.dumps({'detail': status.status_text[status_code]})
    elif status_code == status.HTTP_204_NO_CONTENT or body == b'null':
        body = None

    await monitoring.after(status_code)

    await _http_response_start(send, headers=headers, status_code=status_code)
    await _http_response_body(send, body=body)


def import_class(dotted_path: str, /) -> type:
    """
    Example:
    -------
        Input: panther.db.models.User
        Output: User (The Class)
    """
    path, name = dotted_path.rsplit('.', 1)
    module = importlib.import_module(path)
    return getattr(module, name)


def read_multipart_form_data(boundary: str, body: bytes) -> dict:
    boundary = b'--' + boundary.encode()
    new_line = b'\r\n' if body[-2:] == b'\r\n' else b'\n'

    field_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("'
            + 2 * new_line
            + b')(.*)'
    )
    file_pattern = (
            rb'(Content-Disposition: form-data; name=")(.*)("; filename=")(.*)("'
            + new_line
            + b'Content-Type: )(.*)'
    )

    data = {}
    for _row in body.split(boundary):
        row = _row.removeprefix(new_line).removesuffix(new_line)

        if row in (b'', b'--'):
            continue

        if match := re.match(pattern=field_pattern, string=row):
            _, field_name, _, value = match.groups()
            data[field_name.decode('utf-8')] = value.decode('utf-8')

        else:
            file_meta_data, value = row.split(2 * new_line, 1)

            if match := re.match(pattern=file_pattern, string=file_meta_data):
                _, field_name, _, file_name, _, content_type = match.groups()
                file = File(
                    file_name=file_name.decode('utf-8'),
                    content_type=content_type.decode('utf-8'),
                    file=value,
                )
                data[field_name.decode('utf-8')] = file
            else:
                logger.error('Unrecognized Pattern')

    return data


def generate_ws_connection_id() -> str:
    return uuid4().hex


def is_function_async(func: Callable) -> bool:
    """
    Sync result is 0 --> False
    async result is 128 --> True
    """
    return bool(func.__code__.co_flags & (1 << 7))


def clean_traceback_message(exception: Exception) -> str:
    """We are ignoring packages traceback message"""
    tb = TracebackException(type(exception), exception, exception.__traceback__)
    stack = tb.stack.copy()
    for t in stack:
        if t.filename.find('site-packages/panther') != -1:
            tb.stack.remove(t)
    _traceback = list(tb.format(chain=False))
    return exception if len(_traceback) == 1 else f'{exception}\n' + ''.join(_traceback)


def reformat_code(base_dir):
    try:
        subprocess.run(['ruff', 'format', base_dir])
        subprocess.run(['ruff', 'check', '--select', 'I', '--fix', base_dir])
    except FileNotFoundError:
        raise PantherException("No module named 'ruff', Hint: `pip install ruff`")


def check_function_type_endpoint(endpoint: types.FunctionType) -> Callable:
    # Function Doesn't Have @API Decorator
    if not hasattr(endpoint, '__wrapped__'):
        logger.critical(f'You may have forgotten to use @API() on the {endpoint.__name__}()')
        raise TypeError
    return endpoint


def check_class_type_endpoint(endpoint: Callable) -> Callable:
    from panther.app import GenericAPI

    if not issubclass(endpoint, GenericAPI):
        logger.critical(f'You may have forgotten to inherit from GenericAPI on the {endpoint.__name__}()')
        raise TypeError

    return endpoint.call_method



import asyncio
import random
from pathlib import Path
from unittest import TestCase

import faker
import pytest

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connection import db
from panther.exceptions import DBException

f = faker.Faker()


class Book(Model):
    name: str
    author: str
    pages_count: int


class _BaseDatabaseTestCase:

    # # # Insert
    def test_insert_one(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        book = Book.insert_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count

    def test_insert_many(self):
        insert_count = self._insert_many()
        assert insert_count > 1

    # # # FindOne
    def test_find_one_not_found(self):
        # Insert Many
        self._insert_many()

        # Find One
        book = Book.find_one(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    def test_find_one_in_many_when_its_last(self):
        # Insert Many
        self._insert_many()

        # Insert One
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        created_book = Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Find One
        book = Book.find_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert str(book._id) == str(book.id)
        assert book.name == name
        assert book.pages_count == pages_count
        assert created_book == book

    def test_find_one_in_many_when_its_middle(self):
        # Insert Many
        insert_count = self._insert_many()

        # Insert One
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        created_book = Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Insert Many
        self._insert_many()

        # Find One
        book = Book.find_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count
        assert created_book == book

    def test_first(self):
        # Insert Many
        insert_count = self._insert_many()

        # Insert Many With Same Params
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)

        # Find First
        book = Book.first(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count

    def test_first_not_found(self):
        # Insert Many
        self._insert_many()

        # Find First
        book = Book.first(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    def test_last(self):
        # Insert Many
        self._insert_many()

        # Insert Many With Same Params
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)

        # Find One
        book = Book.last(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book._id == Book.count()
        assert book.name == name
        assert book.pages_count == pages_count

    def test_last_not_found(self):
        self._insert_many()

        # Find Last
        book = Book.last(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    # # # Find
    def test_find(self):
        # Insert Many
        self._insert_many()

        # Insert Many With Specific Name
        name = f.name()
        insert_count = self._insert_many_with_specific_params(name=name)

        # Find
        books = Book.find(name=name)

        assert isinstance(books, list)
        assert len(books) == insert_count
        for book in books:
            assert isinstance(book, Book)
            assert book.name == name

    def test_find_not_found(self):
        # Insert Many
        self._insert_many()

        # Find
        books = Book.find(name='NotFound')
        assert isinstance(books, list)
        assert len(books) == 0

    def test_find_without_filter(self):
        # Insert Many
        insert_count = self._insert_many()

        # Find All
        books = Book.find()

        assert isinstance(books, list)
        assert len(books) == insert_count
        for book in books:
            assert isinstance(book, Book)

    def test_all(self):
        # Insert Many
        insert_count = self._insert_many()

        # Find All
        books = Book.all()

        assert isinstance(books, list)
        assert len(books) == insert_count
        for book in books:
            assert isinstance(book, Book)

    def test_all_not_found(self):
        # Find All
        books = Book.all()

        assert isinstance(books, list)
        assert len(books) == 0

    # # # Count
    def test_count_all(self):
        # Insert Many
        insert_count = self._insert_many()

        # Count All
        books_count = Book.count()

        assert isinstance(books_count, int)
        assert books_count == insert_count

    def test_count_with_filter(self):
        # Insert Many
        self._insert_many()

        # Insert Many With Specific Name
        name = f.name()
        insert_count = self._insert_many_with_specific_params(name=name)

        # Count
        books_count = Book.count(name=name)

        assert isinstance(books_count, int)
        assert books_count == insert_count

    def test_count_not_found(self):
        # Insert Many
        self._insert_many()

        # Count
        books_count = Book.count(name='NotFound')

        assert isinstance(books_count, int)
        assert books_count == 0

    # # # Delete One
    def test_delete_one(self):
        # Insert Many
        self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = self._insert_many_with_specific_params(name=name)

        # Delete One
        is_deleted = Book.delete_one(name=name)

        assert isinstance(is_deleted, bool)
        assert is_deleted is True

        # Count Them After Deletion
        assert Book.count(name=name) == insert_count - 1

    def test_delete_self(self):
        # Insert Many
        self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = self._insert_many_with_specific_params(name=name)
        # Delete One
        book = Book.find_one(name=name)
        book.delete()
        # Count Them After Deletion
        assert Book.count(name=name) == insert_count - 1

    def test_delete_one_not_found(self):
        # Insert Many
        insert_count = self._insert_many()

        # Delete One
        is_deleted = Book.delete_one(name='_Invalid_Name_')

        assert isinstance(is_deleted, bool)
        assert is_deleted is False

        # Count All
        assert Book.count() == insert_count

    # # # Delete Many
    def test_delete_many(self):
        # Insert Many
        pre_insert_count = self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = self._insert_many_with_specific_params(name=name)

        # Delete Many
        deleted_count = Book.delete_many(name=name)

        assert isinstance(deleted_count, int)
        assert deleted_count == insert_count

        # Count Them After Deletion
        assert Book.count(name=name) == 0
        assert Book.count() == pre_insert_count

    def test_delete_many_not_found(self):
        # Insert Many
        pre_insert_count = self._insert_many()

        # Delete Many
        name = 'NotFound'
        deleted_count = Book.delete_many(name=name)

        assert isinstance(deleted_count, int)
        assert deleted_count == 0

        # Count Them After Deletion
        assert Book.count(name=name) == 0
        assert Book.count() == pre_insert_count

    # # # Update
    def test_update_one(self):
        # Insert Many
        insert_count = self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Update One
        new_name = 'New Name'
        is_updated = Book.update_one({'name': name}, name=new_name)

        assert isinstance(is_updated, bool)
        assert is_updated is True

        book = Book.find_one(name=new_name)
        assert isinstance(book, Book)
        assert book.author == author
        assert book.pages_count == pages_count

        # Count Them After Update
        assert Book.count(name=name) == 0
        assert Book.count() == insert_count + 1

    def test_update_self(self):
        # Insert Many
        insert_count = self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Update One
        book = Book.find_one(name=name)
        new_name = 'New Name'
        book.update(name=new_name)

        assert book.name == new_name
        assert book.author == author

        book = Book.find_one(name=new_name)
        assert isinstance(book, Book)
        assert book.author == author
        assert book.pages_count == pages_count

        # Count Them After Update
        assert Book.count(name=name) == 0
        assert Book.count() == insert_count + 1

    def test_update_one_not_found(self):
        # Insert Many
        insert_count = self._insert_many()

        # Update One
        new_name = 'New Name'
        is_updated = Book.update_one({'name': 'NotFound'}, name=new_name)

        assert isinstance(is_updated, bool)
        assert is_updated is False

        book = Book.find_one(name=new_name)
        assert book is None

        # Count Them After Update
        assert Book.count() == insert_count

    # # # Update Many
    def test_update_many(self):
        # Insert Many
        pre_insert_count = self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        insert_count = self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)

        # Update Many
        new_name = 'New Name'
        updated_count = Book.update_many({'name': name}, name=new_name)

        assert isinstance(updated_count, int)
        assert updated_count == insert_count

        books = Book.find(name=new_name)
        assert isinstance(books, list)
        assert len(books) == updated_count == insert_count
        for book in books:
            assert book.author == author
            assert book.pages_count == pages_count

        # Count Them After Update
        assert Book.count() == pre_insert_count + insert_count

    def test_update_many_not_found(self):
        # Insert Many
        insert_count = self._insert_many()

        # Update Many
        new_name = 'New Name'
        updated_count = Book.update_many({'name': 'NotFound'}, name=new_name)

        assert isinstance(updated_count, int)
        assert updated_count == 0

        book = Book.find_one(name=new_name)
        assert book is None

        # Count Them After Update
        assert Book.count() == insert_count

    @classmethod
    def _insert_many(cls) -> int:
        insert_count = random.randint(2, 10)

        for _ in range(insert_count):
            Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        return insert_count

    @classmethod
    def _insert_many_with_specific_params(
            cls,
            name: str = f.name(),
            author: str = f.name(),
            pages_count: int = random.randint(0, 10)
    ) -> int:
        insert_count = random.randint(2, 10)

        for _ in range(insert_count):
            Book.insert_one(name=name, author=author, pages_count=pages_count)

        return insert_count


class TestPantherDB(_BaseDatabaseTestCase, TestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global MIDDLEWARES
        MIDDLEWARES = [
            ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{cls.DB_PATH}'}),
        ]
        Panther(__name__, configs=__name__, urls={})

    def setUp(self) -> None:
        for middleware in config['http_middlewares']:
            asyncio.run(middleware.before(request=None))

    def tearDown(self) -> None:
        Path(self.DB_PATH).unlink()
        for middleware in config['reversed_http_middlewares']:
            asyncio.run(middleware.after(response=None))


@pytest.mark.mongodb
class TestMongoDB(_BaseDatabaseTestCase, TestCase):
    DB_NAME = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global MIDDLEWARES
        MIDDLEWARES = [
            ('panther.middlewares.db.DatabaseMiddleware', {'url': f'mongodb://127.0.0.1:27017/{cls.DB_NAME}'}),
        ]
        Panther(__name__, configs=__name__, urls={})

    def setUp(self) -> None:
        for middleware in config['http_middlewares']:
            asyncio.run(middleware.before(request=None))

    def tearDown(self) -> None:
        db.session.drop_collection('Book')
        for middleware in config['reversed_http_middlewares']:
            asyncio.run(middleware.after(response=None))

    def test_last(self):
        try:
            super().test_last()
        except DBException as exc:
            assert exc.args[0] == 'last() is not supported in MongoDB yet.'
        else:
            assert False

    def test_last_not_found(self):
        try:
            super().test_last_not_found()
        except DBException as exc:
            assert exc.args[0] == 'last() is not supported in MongoDB yet.'
        else:
            assert False



import time
from unittest import TestCase

from panther.background_tasks import BackgroundTasks, BackgroundTask
from panther.configs import config
from panther.utils import Singleton


class TestBackgroundTasks(TestCase):
    def setUp(self):
        self.obj = BackgroundTasks()
        config['background_tasks'] = True

    def tearDown(self):
        del Singleton._instances[BackgroundTasks]
        config['background_tasks'] = False

    def test_background_tasks_singleton(self):
        new_obj = BackgroundTasks()
        self.assertEqual(id(self.obj), id(new_obj))

    def test_initialization(self):
        self.assertTrue(hasattr(self.obj, 'initialize'))
        self.assertFalse(self.obj._initialized)

        self.obj.initialize()
        self.assertTrue(self.obj._initialized)
        self.assertEqual(self.obj.tasks, [])

    def test_add_single_task(self):
        def func():
            pass

        task = BackgroundTask(func)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])

    def test_add_wrong_task(self):
        def func(): pass

        self.obj.initialize()
        self.assertEqual(self.obj.tasks, [])

        with self.assertLogs() as captured:
            self.obj.add_task(func)

        self.assertEqual(len(captured.records), 1)
        self.assertEqual(
            captured.records[0].getMessage(),
            f'`{func.__name__}` should be instance of `background_tasks.BackgroundTask`'
        )
        self.assertEqual(self.obj.tasks, [])

    def test_add_task_with_false_background_task(self):
        numbers = []

        def func(_numbers): _numbers.append(1)

        task = BackgroundTask(func, numbers)
        with self.assertLogs() as captured:
            self.obj.add_task(task)
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(
            captured.records[0].getMessage(),
            'Task will be ignored, `BACKGROUND_TASKS` is not True in `core/configs.py`'
        )
        self.assertEqual(self.obj.tasks, [])

    def test_add_task_with_args(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])
        time.sleep(2)
        self.assertEqual(len(numbers), 1)

    def test_add_task_with_kwargs(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, _numbers=numbers)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])
        time.sleep(2)
        self.assertEqual(len(numbers), 1)

    def test_add_task_with_custom_interval(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).every_seconds().interval(3)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])
        time.sleep(4)
        self.assertEqual(len(numbers), 3)

    def test_add_task_with_custom_interval_default_timedelta(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).interval(2)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])
        time.sleep(3)
        # default timedelta is 1 minute, so it can't complete the intervals in 2 seconds
        self.assertEqual(len(numbers), 1)

    def test_add_multiple_tasks(self):
        numbers = []

        def func1(_numbers):
            _numbers.append(1)

        def func2(_numbers):
            _numbers.append(1)

        task2 = BackgroundTask(func1, numbers)
        task1 = BackgroundTask(func2, numbers)
        self.obj.initialize()
        self.obj.add_task(task1)
        self.obj.add_task(task2)
        self.assertEqual(self.obj.tasks, [task1, task2])
        self.assertEqual(len(numbers), 0)
        time.sleep(2)
        self.assertEqual(len(numbers), 2)

    def test_add_task_with_custom_every_seconds(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).every_seconds(3).interval(2)
        self.obj.initialize()
        self.obj.add_task(task)
        self.assertEqual(self.obj.tasks, [task])
        time.sleep(3)
        self.assertEqual(len(numbers), 1)
        time.sleep(3)
        self.assertEqual(len(numbers), 2)




from pathlib import Path
from unittest import TestCase

from pydantic import Field, ConfigDict
from pydantic import field_validator

from panther import Panther
from panther.app import API
from panther.db import Model
from panther.request import Request
from panther.serializer import ModelSerializer
from panther.test import APIClient


class Book(Model):
    name: str
    author: str = Field('default_author')
    pages_count: int = Field(0)


class NotRequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['author', 'pages_count']


class RequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']


class OnlyRequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']


class WithValidatorsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']

    @field_validator('name', 'author', 'pages_count')
    def validate(cls, field):
        return 'validated'


class WithClassFieldsSerializer(ModelSerializer):
    age: int = Field(10)

    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']


@API(input_model=NotRequiredFieldsSerializer)
async def not_required(request: Request):
    return request.validated_data


@API(input_model=RequiredFieldsSerializer)
async def required(request: Request):
    return request.validated_data


@API(input_model=OnlyRequiredFieldsSerializer)
async def only_required(request: Request):
    return request.validated_data


@API(input_model=WithValidatorsSerializer)
async def with_validators(request: Request):
    return request.validated_data


@API(input_model=WithClassFieldsSerializer)
async def with_class_fields(request: Request):
    return request.validated_data


urls = {
    'not-required': not_required,
    'required': required,
    'only-required': only_required,
    'with-validators': with_validators,
    'class-fields': with_class_fields,
}


class TestModelSerializer(TestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global MIDDLEWARES
        MIDDLEWARES = [
            ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{cls.DB_PATH}'}),
        ]
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def tearDown(self) -> None:
        Path(self.DB_PATH).unlink(missing_ok=True)

    # # # Class Usage

    def test_not_required_fields_empty_response(self):
        payload = {}
        res = self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'default_author', 'pages_count': 0}

    def test_not_required_fields_full_response(self):
        payload = {
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'ali', 'pages_count': 12}

    def test_required_fields_error(self):
        payload = {}
        res = self.client.post('required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required'}

    def test_required_fields_success(self):
        payload = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    def test_only_required_fields_error(self):
        payload = {}
        res = self.client.post('only-required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required', 'author': 'Field required', 'pages_count': 'Field required'}

    def test_only_required_fields_success(self):
        payload = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('only-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    def test_with_validators(self):
        payload = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('with-validators', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'validated', 'author': 'validated', 'pages_count': 'validated'}

    def test_with_class_fields_success(self):
        # Test Default Value
        payload1 = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('class-fields', payload=payload1)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12, 'age': 10}

        # Test Validation
        payload2 = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12',
            'age': 30
        }
        res = self.client.post('class-fields', payload=payload2)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12, 'age': 30}

    # # # Class Definition

    def test_define_class_without_meta(self):
        try:
            class Serializer0(ModelSerializer):
                pass
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`class Config` is required in tests.test_serializer.Serializer0.'
        else:
            assert False

    def test_define_class_without_model(self):
        try:
            class Serializer1(ModelSerializer):
                class Config:
                    pass
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer1.Config.model` is required.'
        else:
            assert False

    def test_define_class_without_fields(self):
        try:
            class Serializer2(ModelSerializer):
                class Config:
                    model = Book
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer2.Config.fields` is required.'
        else:
            assert False

    def test_define_class_with_invalid_fields(self):
        try:
            class Serializer3(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['ok', 'no']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer3.Config.fields.ok` is not valid.'
        else:
            assert False

    def test_define_class_with_invalid_required_fields(self):
        try:
            class Serializer4(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['name', 'author']
                    required_fields = ['pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer4.Config.required_fields.pages_count` should be in `Config.fields` too.'
        else:
            assert False

    def test_define_class_with_invalid_model(self):
        try:
            class Serializer5(ModelSerializer):
                class Config:
                    model = ModelSerializer
                    fields = ['name', 'author', 'pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer5.Config.model` is not subclass of `panther.db.Model`.'
        else:
            assert False

    # # # Serializer Usage
    def test_with_simple_model_config(self):
        class Serializer(ModelSerializer):
            model_config = ConfigDict(str_to_upper=True)

            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'BOOK'
        assert serialized.author == 'ALIRN'
        assert serialized.pages_count == 12

    def test_with_inner_model_config(self):
        class Serializer(ModelSerializer):
            class Config:
                str_to_upper = True
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'BOOK'
        assert serialized.author == 'ALIRN'
        assert serialized.pages_count == 12

    def test_with_dual_model_config(self):
        class Serializer(ModelSerializer):
            model_config = ConfigDict(str_to_upper=False)

            class Config:
                str_to_upper = True
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'book'
        assert serialized.author == 'AliRn'
        assert serialized.pages_count == 12

    def test_serializer_doc(self):
        class Serializer1(ModelSerializer):
            """Hello I'm Doc"""
            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer1(name='book', author='AliRn', pages_count='12')
        assert serialized.__doc__ == 'Hello I\'m Doc'

        class Serializer2(ModelSerializer):
            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer2(name='book', author='AliRn', pages_count='12')
        assert serialized.__doc__ is None



import os
import shutil
import sys
from io import StringIO
from pathlib import Path
from unittest import TestCase, skipIf
from unittest.mock import patch

from rich import print as rprint

from panther import Panther
from panther.cli.create_command import CreateProject
from panther.cli.template import TEMPLATE, SINGLE_FILE_TEMPLATE

interactive_cli_1_index = 0
interactive_cli_2_index = 0


def interactive_cli_1_mock_responses(index=None):
    global interactive_cli_1_index
    if index is None:
        index = interactive_cli_1_index
    responses = ['project1', 'project1_dir', 'n', '0', 'y', 'y', 'y', 'y', 'y']
    response = responses[index]
    interactive_cli_1_index += 1
    return response


def interactive_cli_2_mock_responses(index=None):
    global interactive_cli_2_index
    if index is None:
        index = interactive_cli_2_index
    responses = ['project2', 'project2_dir', 'y', '0', 'y', 'y', 'y', 'y', 'y']
    response = responses[index]
    interactive_cli_2_index += 1
    return response


class TestCLI(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/sample_project')

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    @skipIf(sys.platform.startswith('win'), 'Not supported in windows')
    def test_print_info(self):
        with patch('sys.stdout', new=StringIO()) as fake_out1:
            app = Panther(__name__)

        base_dir = '{0:<39}'.format(str(Path(__name__).absolute().parent))
        expected_value = fr"""╭──────────────────────────────────────────────────────────╮
│    ____                 __    __                         │
│   /\  _`\              /\ \__/\ \                        │
│   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     │
│    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   │
│     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    │
│      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    │
│       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    │
│                                                          │
│   Monitoring: True                                       │
│   Log Queries: True                                      │
│   Background Tasks: False                                │
│   Websocket: False                                       │
│   Base directory: {base_dir}│
│ * Run "panther monitor" in another session for Monitoring│
│ * You may want to install `uvloop` for better performance│
│   `pip install uvloop`                                   │
╰──────────────────────────────────────────────────────────╯"""
        with patch('sys.stdout', new=StringIO()) as fake_out2:
            rprint(expected_value)
        assert fake_out1.getvalue() == fake_out2.getvalue()

    @patch('builtins.input', interactive_cli_1_mock_responses)
    def test_create_normal_template_with_interactive_cli(self):
        CreateProject().create([])

        project_path = interactive_cli_1_mock_responses(1)
        for file_name, data in SINGLE_FILE_TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()
        shutil.rmtree(project_path)

    @patch('builtins.input', interactive_cli_2_mock_responses)
    def test_create_single_file_template_with_interactive_cli(self):
        CreateProject().create([])

        project_path = interactive_cli_2_mock_responses(1)
        for file_name, data in SINGLE_FILE_TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()
        shutil.rmtree(project_path)

    def test_create_on_existence_directory(self):
        project_path = 'test-project-directory'
        os.mkdir(project_path)

        with self.assertLogs(level='ERROR') as captured_error:
            CreateProject().create(['test_project', project_path])

        with self.assertLogs(level='INFO') as captured_info:
            CreateProject().create(['test_project', project_path])

        try:
            assert len(captured_error.records) == 1
            assert captured_info.records[0].getMessage() == f'"{project_path}" Directory Already Exists.'

            assert len(captured_info.records) == 2
            assert captured_info.records[0].getMessage() == f'"{project_path}" Directory Already Exists.'
            assert captured_info.records[1].getMessage() == 'Use "panther -h" for more help'
        except AssertionError as e:
            raise e
        finally:
            os.removedirs(project_path)

    def test_create_project(self):
        project_path = 'test-project-directory'
        CreateProject().create(['test_project', project_path])

        for file_name, data in TEMPLATE.items():
            sub_directory = f'{project_path}/{file_name}'
            assert Path(sub_directory).exists()

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    assert Path(file_path).exists()

        shutil.rmtree(project_path)



import contextlib
import os
from collections import deque
from pathlib import Path

from rich import box
from rich.align import Align
from rich.console import Group
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from watchfiles import watch

from panther.cli.utils import cli_error
from panther.configs import config


def monitor() -> None:
    monitoring_log_file = Path(config['base_dir'] / 'logs' / 'monitoring.log')

    def _generate_table(rows: deque) -> Panel:
        layout = Layout()

        rows = list(rows)
        _, lines = os.get_terminal_size()

        table = Table(box=box.MINIMAL_DOUBLE_HEAD)
        table.add_column('Datetime', justify='center', style='magenta', no_wrap=True)
        table.add_column('Method', justify='center', style='cyan')
        table.add_column('Path', justify='center', style='cyan')
        table.add_column('Client', justify='center', style='cyan')
        table.add_column('Response Time', justify='center', style='blue')
        table.add_column('Status Code', justify='center', style='blue')

        for row in rows[-lines:]:  # It will give us "lines" last lines of "rows"
            table.add_row(*row)
        layout.update(table)

        return Panel(
            Align.center(Group(table)),
            box=box.ROUNDED,
            padding=(1, 2),
            title='Monitoring',
            border_style='bright_blue',
        )

    if not monitoring_log_file.exists():
        return cli_error('Monitoring file not found. (You need at least one monitoring record for this action)')

    with monitoring_log_file.open() as f:
        f.readlines()  # Set cursor at the end of file

        _, init_lines_count = os.get_terminal_size()
        messages = deque(maxlen=init_lines_count - 10)  # Save space for header and footer

        with (
            Live(
                _generate_table(messages),
                auto_refresh=False,
                vertical_overflow='visible',
                screen=True,
            ) as live,
            contextlib.suppress(KeyboardInterrupt),
        ):
            for _ in watch(monitoring_log_file):
                data = f.readline().split('|')
                # 2023-03-24 01:42:52 | GET | /user/317/ | 127.0.0.1:48856 |  0.0366 ms | 200
                messages.append(data)
                live.update(_generate_table(messages))
                live.refresh()



from panther.request import Request
from panther.response import Response
from panther.websocket import GenericWebsocket


class BaseMiddleware:
    async def before(self, request: Request | GenericWebsocket):
        raise NotImplementedError

    async def after(self, response: Response | GenericWebsocket):
        raise NotImplementedError


class HTTPMiddleware(BaseMiddleware):
    async def before(self, request: Request):
        return request

    async def after(self, response: Response):
        return response


class WebsocketMiddleware(BaseMiddleware):
    async def before(self, request: GenericWebsocket):
        return request

    async def after(self, response: GenericWebsocket):
        return response



from unittest import TestCase

from panther import Panther
from panther.app import API
from panther.response import Response
from panther.test import APIClient


@API()
async def return_nothing():
    pass


@API()
async def return_none():
    return None


@API()
async def return_dict():
    return {'detail': 'ok'}


@API()
async def return_list():
    return [1, 2, 3]


@API()
async def return_tuple():
    return 1, 2, 3, 4


@API()
async def return_response_none():
    return Response()


@API()
async def return_response_dict():
    return Response(data={'detail': 'ok'})


@API()
async def return_response_list():
    return Response(data=['car', 'home', 'phone'])


@API()
async def return_response_tuple():
    return Response(data=('car', 'home', 'phone', 'book'))


urls = {
    'nothing': return_nothing,
    'none': return_none,
    'dict': return_dict,
    'list': return_list,
    'tuple': return_tuple,
    'response-none': return_response_none,
    'response-dict': return_response_dict,
    'response-list': return_response_list,
    'response-tuple': return_response_tuple,
}


class TestSimpleResponses(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def test_nothing(self):
        res = self.client.get('nothing/')
        assert res.status_code == 200
        assert res.data is None

    def test_none(self):
        res = self.client.get('none/')
        assert res.status_code == 200
        assert res.data is None

    def test_dict(self):
        res = self.client.get('dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}

    def test_list(self):
        res = self.client.get('list/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3]

    def test_tuple(self):
        res = self.client.get('tuple/')
        assert res.status_code == 200
        assert res.data == [1, 2, 3, 4]

    def test_response_none(self):
        res = self.client.get('response-none/')
        assert res.status_code == 200
        assert res.data is None

    def test_response_dict(self):
        res = self.client.get('response-dict/')
        assert res.status_code == 200
        assert res.data == {'detail': 'ok'}

    def test_response_list(self):
        res = self.client.get('response-list/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone']

    def test_response_tuple(self):
        res = self.client.get('response-tuple/')
        assert res.status_code == 200
        assert res.data == ['car', 'home', 'phone', 'book']



import asyncio
from pathlib import Path
from unittest import TestCase

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.db import Model
from panther.request import Request
from panther.test import APIClient


@API()
async def without_auth(request: Request):
    return request.user


@API(auth=True)
async def auth_required(request: Request):
    return request.user


urls = {
    'without': without_auth,
    'auth-required': auth_required,
}


class User(Model):
    username: str
    password: str


AUTHENTICATION = 'panther.authentications.JWTAuthentication'
SECRET_KEY = 'hvdhRspoTPh1cJVBHcuingQeOKNc1uRhIP2k7suLe2g='
DB_PATH = 'test.pdb'
MIDDLEWARES = [
    ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{DB_PATH}'}),
]
USER_MODEL = 'tests.test_authentication.User'


class TestAuthentication(TestCase):
    SHORT_TOKEN = {'Authorization': 'TOKEN'}
    NOT_ENOUGH_SEGMENT_TOKEN = {'Authorization': 'Bearer XXX'}
    JUST_BEARER_TOKEN = {'Authorization': 'Bearer'}
    BAD_UNICODE_TOKEN = {'Authorization': 'Bearer علی'}
    BAD_SIGNATURE_TOKEN = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.JAWUkAU2mWhxcd6MS8r9pd44yBIfkEBmpr3WLeqIccM'}
    TOKEN_WITHOUT_USER_ID = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJpZCI6MX0.PpyXW0PgmGSPaaNirm_Ei4Y2fw9nb4TN26RN1u9RHSo'}
    TOKEN = {'Authorization': 'Bearer eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyX2lkIjoxfQ.AF3nsj8IQ6t0ncqIx4quoyPfYaZ-pqUOW4z_euUztPM'}

    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def setUp(self) -> None:
        for middleware in config['http_middlewares']:
            asyncio.run(middleware.before(request=None))

    def tearDown(self) -> None:
        for middleware in config['reversed_http_middlewares']:
            asyncio.run(middleware.after(response=None))
        Path(DB_PATH).unlink()

    def test_user_without_auth(self):
        res = self.client.get('without')
        assert res.status_code == 200
        assert res.data is None

        res = self.client.get('without', headers={'Authorization': 'Token'})
        assert res.status_code == 200
        assert res.data is None

    def test_user_auth_required_without_auth_class(self):
        auth_config = config['authentication']
        config['authentication'] = None
        with self.assertLogs(level='CRITICAL') as captured:
            res = self.client.get('auth-required')
        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == '"AUTHENTICATION" has not been set in configs'
        assert res.status_code == 500
        assert res.data['detail'] == 'Internal Server Error'
        config['authentication'] = auth_config

    def test_user_auth_required_without_token(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required')

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization is required"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_bad_token_1(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.SHORT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization keyword is not valid"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_bad_token2(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.JUST_BEARER_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Authorization should have 2 part"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_bad_token3(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.BAD_UNICODE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == (
            'JWT Authentication Error: "\'latin-1\' codec can\'t encode characters in position 7-9: '
            'ordinal not in range(256)"'
        )
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_bad_token4(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.NOT_ENOUGH_SEGMENT_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Not enough segments"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_invalid_token_signature(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.BAD_SIGNATURE_TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Signature verification failed."'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_token_without_user_id(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.TOKEN_WITHOUT_USER_ID)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "Payload does not have user_id"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_token_user_not_found(self):
        with self.assertLogs(level='ERROR') as captured:
            res = self.client.get('auth-required', headers=self.TOKEN)

        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'JWT Authentication Error: "User not found"'
        assert res.status_code == 401
        assert res.data['detail'] == 'Authentication Error'

    def test_user_auth_required_with_token(self):
        User.insert_one(username='Username', password='Password')

        with self.assertNoLogs(level='ERROR'):
            res = self.client.get('auth-required', headers=self.TOKEN)

        assert res.status_code == 200
        assert [*res.data.keys()] == ['id', 'username', 'password']
        assert res.data['id'] == '1'
        assert res.data['username'] == 'Username'
        assert res.data['password'] == 'Password'



import sys
from datetime import timedelta
from pathlib import Path
from unittest import TestCase

import tests.sample_project.app.models
from panther import Panther
from panther.configs import Config


class TestRun(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/sample_project')

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_init(self):
        app = Panther(__name__)
        self.assertIsInstance(app, Panther)

    def test_load_configs(self):
        from panther.configs import config
        from panther.panel.apis import documents_api, models_api, single_document_api

        base_dir = Path(__name__).resolve().parent
        secret_key = 'fHrIYx3yK0J_UG0K0zD6miLPNy1esoYXzVsvif6e7rY='
        Panther(__name__)

        assert isinstance(config, Config)
        assert config['base_dir'] == base_dir
        assert config['monitoring'] is True
        assert config['log_queries'] is True
        assert config['default_cache_exp'] == timedelta(seconds=10)
        assert config['throttling'].rate == 10
        assert config['throttling'].duration == timedelta(seconds=10)
        assert config['secret_key'] == secret_key.encode()

        assert len(config['http_middlewares']) == 1
        assert config['http_middlewares'][0].__class__.__name__ == 'DatabaseMiddleware'
        assert config['http_middlewares'][0].url == 'pantherdb://test.pdb'

        assert len(config['reversed_http_middlewares']) == 1
        assert config['reversed_http_middlewares'][0].__class__.__name__ == 'DatabaseMiddleware'
        assert config['reversed_http_middlewares'][0].url == 'pantherdb://test.pdb'

        assert len(config['ws_middlewares']) == 1
        assert config['ws_middlewares'][0].__class__.__name__ == 'DatabaseMiddleware'
        assert config['ws_middlewares'][0].url == 'pantherdb://test.pdb'

        assert len(config['reversed_ws_middlewares']) == 1
        assert config['reversed_ws_middlewares'][0].__class__.__name__ == 'DatabaseMiddleware'
        assert config['reversed_ws_middlewares'][0].url == 'pantherdb://test.pdb'

        assert config['user_model'].__name__ == tests.sample_project.app.models.User.__name__
        assert config['user_model'].__module__.endswith('app.models')
        assert config['jwt_config'].algorithm == 'HS256'
        assert config['jwt_config'].life_time == timedelta(days=2).total_seconds()
        assert config['jwt_config'].key == secret_key

        assert '' in config['urls']
        config['urls'].pop('')

        assert 'second' in config['urls']
        config['urls'].pop('second')

        urls = {
            '_panel': {
                '': models_api,
                '<index>': {
                    '': documents_api,
                    '<document_id>': single_document_api,
                },
            },
        }
        assert config['urls'] == urls
        assert config['query_engine'].__name__ == 'BasePantherDBQuery'



import re

from setuptools import setup


def panther_version() -> str:
    with open('panther/__init__.py') as f:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", f.read()).group(1)


VERSION = panther_version()
with open('README.md', encoding='utf-8') as file:
    DESCRIPTION = file.read()

EXTRAS_REQUIRE = {
    'full': [
        'redis==5.0.1',
        'pymongo~=4.4',
        'bpython~=0.24',
        'ruff~=0.1.9',
        'python-jose~=3.3',
        'websockets~=12.0',
        'cryptography~=41.0',
    ],
}

setup(
    name='panther',
    version=VERSION,
    python_requires='>=3.10',
    author='Ali RajabNezhad',
    author_email='alirn76@yahoo.com',
    url='https://github.com/alirn76/panther',
    description='Fast &  Friendly, Web Framework For Building Async APIs',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    include_package_data=True,
    license='BSD-3-Clause license',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    entry_points={
        'console_scripts': ['panther=panther.cli.main:start'],
    },
    package_data={
        'panther': ['cli/*'],
    },
    install_requires=[
        'bson~=0.5',
        'httptools~=0.6',
        'pantherdb==1.3.5',
        'pydantic~=2.1',
        'rich~=13.5',
        'uvicorn~=0.23',
        'watchfiles~=0.19',
    ],
    extras_require=EXTRAS_REQUIRE,
)
